/*********************************************************************
Project	:	MacPerl			-	Real Perl Application
File		:	MPFile.c			-
Author	:	Matthias Neeracher

A lot of this code is borrowed from 7Edit written by
Apple Developer Support UK

Language	:	MPW C

$Log: MPFile.c,v $
Revision 1.1  1994/02/27  23:00:47  neeri
Initial revision

Revision 0.6  1993/09/17  00:00:00  neeri
Runtime version

Revision 0.5  1993/08/28  00:00:00  neeri
Handle multiple preference files

Revision 0.4  1993/08/17  00:00:00  neeri
Enable Save

Revision 0.3  1993/08/13  00:00:00  neeri
Write bounds rectangles

Revision 0.2  1993/05/30  00:00:00  neeri
Support Console Windows

Revision 0.1  1993/05/29  00:00:00  neeri
Compiles correctly

*********************************************************************/

#include <Errors.h>
#include <Resources.h>
#include <Desk.h>
#include <PLStringFuncs.h>
#include <AppleEvents.h>
#include <AERegistry.h>
#include <StandardFile.h>
#include <TFileSpec.h>
#include <Balloons.h>

#include "MPFile.h"
#include "MPSave.h"

/**-----------------------------------------------------------------------
		Name: 			FileError
		Purpose:		Puts up an error alert.
	-----------------------------------------------------------------------**/


#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

pascal void FileError(Str255 s, Str255 f)
{
	short    alertResult;

	SetCursor(&qd.arrow);
	ParamText(s, f, (StringPtr) "\p", (StringPtr) "\p");
 	alertResult = Alert(ErrorAlert, nil);
}

/**-----------------------------------------------------------------------
		Name: 			DoClose
		Purpose:		Closes a window.
	-----------------------------------------------------------------------**/

#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

pascal OSErr DoClose(WindowPtr aWindow, Boolean canInteract, DescType dialogAnswer)
{
 	DPtr    aDocument;
 	short   alertResult;
 	Str255  theName;
	OSErr   myErr;

	myErr = noErr;

 	if (gWCount > 0) {
		aDocument = DPtrFromWindowPtr(aWindow);

		if (aDocument->kind == kDocumentWindow) {
			if (aDocument->dirty)
				if (canInteract && (dialogAnswer==kAEAsk)) {
					if (aDocument->u.reg.everSaved == false)
						GetWTitle(aWindow, theName); /* Pick it up as a script may have changed it */
					else
						PLstrcpy(theName, aDocument->theFileName);
	
					ParamText(theName, (StringPtr) "\p", (StringPtr) "\p", (StringPtr) "\p");
					SetCursor(&qd.arrow);
					alertResult = Alert(SaveAlert, nil);
					switch (alertResult) {
					case aaSave:
						myErr = SaveAskingName(aDocument, canInteract);
						break;
	
					case aaCancel:
						return userCanceledErr;
	
					case aaDiscard:
						aDocument->dirty = false;
						break;
					}
				} else {
					if (dialogAnswer==kAEYes)
						myErr = SaveAskingName(aDocument, canInteract);
					else
						myErr = noErr; /* Don't save */
				}
				
			if (!myErr) {
#ifndef RUNTIME
				if (aDocument->kind == kDocumentWindow && aDocument->u.reg.numSections)
					DeRegisterAllSections(aDocument);
#endif

				CloseMyWindow(aWindow);
			}
		} else if (aDocument->u.cons.selected) {
			if (!gQuitting) {
				SysBeep(0);
	
				return userCanceledErr;
			}
		} else 
			SaveConsole(aDocument);
	} else
		myErr = errAEIllegalIndex;

	return myErr;
}

#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

//  DoQuit
//  saveOpt - one of kAEAsk,kAEYes,kAENo
//  if kAEYes or kAEAsk then AEInteactWithUser should have been called
//  before DoQuit. Assumes that it can interact if it needs to.

pascal void DoQuit(DescType saveOpt)
{
	WindowPeek	aWindow;
	WindowPeek	nextWindow;
	short			theKind;

	if (gRunningPerl && (Alert(AbortAlert, nil) == 2))
		return;
		
	gQuitting = true;

	for (aWindow = (WindowPeek) FrontWindow(); aWindow; aWindow = nextWindow) {
		nextWindow = aWindow->nextWindow;
		if (Ours((WindowPtr) aWindow)) {
			if (DoClose((WindowPtr) aWindow, true, saveOpt)) {
				gQuitting = false;
				
				return;
			}
		} else {
			theKind = aWindow->windowKind;
			if (theKind < 0)
				CloseDeskAcc(theKind);
		}
	}
}

#ifndef RUNTIME
pascal Boolean GetFileFilter(CInfoPBPtr pb)
{
	switch (GetDocTypeFromInfo(pb)) {
	case kPreferenceDoc:
		/* We don't want preference files here. Maybe we should */
	case kUnknownDoc:
		return true;
	default:
		return false;
	}
}

#if USESROUTINEDESCRIPTORS
RoutineDescriptor	uGetFileFilter = 
		BUILD_ROUTINE_DESCRIPTOR(uppFileFilterProcInfo, GetFileFilter);
#else
#define uGetFileFilter *(FileFilterUPP)&GetFileFilter
#endif

#else
pascal Boolean GetFileFilter(ParmBlkPtr info)
{
	switch (info->fileParam.ioFlFndrInfo.fdType) {
	case 'APPL':
		switch (info->fileParam.ioFlFndrInfo.fdCreator) {
		case MPRtSig:
			return false;
		case MPAppSig:
			return !info->fileParam.ioFlLgLen;
		default:
			return true;
		}
	case 'TEXT':
		return false;
	default:
		return true;
	}
}
#endif

pascal OSErr GetFile(FSSpec *theFSSpec)
{
#ifndef RUNTIME
	StandardFileReply  reply;

	BuildSEList();
	
	StandardGetFile(&uGetFileFilter, MacPerlFileTypeCount, MacPerlFileTypes, &reply);

	if (reply.sfGood) {
		*theFSSpec = reply.sfFile;
		return noErr;
	} else
		return userCanceledErr;
		
#else

	SFTypeList 	myTypes = {'TEXT', 'APPL'};
	SFReply	  	reply;
	Point			where;

	where.h = where.v = 75;

	SFGetFile(
		where, (StringPtr) "\p", 
		(FileFilterUPP) GetFileFilter, 2, myTypes, 
		(DlgHookProcPtr) nil,
		&reply);

	if (reply.good)
		return WD2FSSpec(reply.vRefNum, reply.fName, theFSSpec);
	else
		return userCanceledErr;
		
#endif
}

#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

pascal OSErr DoCreate(FSSpec theSpec)
{
	OSErr err;

#ifndef RUNTIME
	err = FSpCreate(&theSpec, MPAppSig, 'TEXT', smSystemScript);
#else
	err = HCreate(theSpec.vRefNum, theSpec.parID, theSpec.name, MPAppSig, 'TEXT');
#endif

	if (!err)
		HCreateResFile(theSpec.vRefNum, theSpec.parID, theSpec.name);
	else
		ShowError((StringPtr) "\pCreating", err);

	return err;
}

pascal OSErr SaveConsole(DPtr doc)
{
	OSErr			err;
	short       resFile;
	HHandle     theHHandle;
	Str255		title;
	Boolean		existing;
	
	if (gPrefsFile) {
		resFile	=	CurResFile();
		UseResFile(gPrefsFile);
		GetWTitle(doc->theWindow, title);
		
		if (theHHandle = (HHandle) Get1NamedResource('TFSS', title)) {
			existing = true;
		} else {
			existing = false;
			theHHandle = (HHandle)NewHandle(sizeof(HeaderRec));
		}
		
		HLock((Handle)theHHandle);
	
		(*theHHandle)->theRect     = doc->theWindow->portRect;
		OffsetRect(
			&(*theHHandle)->theRect,
			-doc->theWindow->portBits.bounds.left,
			-doc->theWindow->portBits.bounds.top);
			
		GetFontName((*(doc->theText))->txFont, (StringPtr) &(*theHHandle)->theFont);
		
		(*theHHandle)->theSize     = (*(doc->theText))->txSize;
		(*theHHandle)->lastID      = 0;
		(*theHHandle)->numSections = 0;
	
		HUnlock((Handle)theHHandle);
	
		if (existing) {
			ChangedResource((Handle) theHHandle);
			WriteResource((Handle) theHHandle);
			UpdateResFile(gPrefsFile);
		} else {
			AddResource((Handle)theHHandle, 'TFSS', Unique1ID('TFSS'), title);
		}
		
		err = ResError();
		
		UseResFile(resFile);
	}
	
	if (doc->u.cons.cookie) {
		/* We might need this window again. */
		DoHideWindow(doc->theWindow);
		TESetSelect(0, 32767, doc->theText);
		TEDelete(doc->theText);
	
		if (doc->u.cons.fence < 32767)
			doc->u.cons.fence	= 0;
	} else /* Done with this window */
		CloseMyWindow(doc->theWindow);
	
	return err;
} 

pascal void ApplySettings(DPtr doc, HPtr settings)
{
	short		fNum;
	FontInfo	info;
	Rect		bounds;
	
	GetFNum(settings->theFont, &fNum);
	SetPort(doc->theWindow);
	TextFont(fNum);
	TextSize(settings->theSize);
	GetFontInfo(&info);
	
	(*doc->theText)->txFont 		= fNum;
	(*doc->theText)->txSize 		= settings->theSize;
	(*doc->theText)->lineHeight	= info.ascent+info.descent+info.leading;
	(*doc->theText)->fontAscent	= info.ascent;
	
	bounds.top							= settings->theRect.top - 13;
	bounds.left							= settings->theRect.left + 5;
	bounds.bottom						= settings->theRect.top  - 5;
	bounds.right						= settings->theRect.right - 5;
	
	if (settings->theRect.right > settings->theRect.left + 50 &&
		settings->theRect.bottom > settings->theRect.top  + 50 &&
		RectInRgn(&bounds, GetGrayRgn())
	) {
		MoveWindow(doc->theWindow, settings->theRect.left, settings->theRect.top, false);
		SizeWindow(
			doc->theWindow,
			settings->theRect.right - settings->theRect.left,
			settings->theRect.bottom - settings->theRect.top,
			false);
	}
		
	ResizeWindow(doc);
}

pascal void RestoreConsole(DPtr doc)
{
	short       resFile;
	HHandle     theHHandle;
	Str255		title;
	
	if (!gPrefsFile)
		return;
		
	resFile	=	CurResFile();
	UseResFile(gPrefsFile);
	GetWTitle(doc->theWindow, title);
	
	if (theHHandle = (HHandle) Get1NamedResource('TFSS', title)) {
 		HLock((Handle)theHHandle);
		
		ApplySettings(doc, *theHHandle);
		
		HUnlock((Handle)theHHandle);
	}
	
	UseResFile(resFile);
}


/** -----------------------------------------------------------------------
		Name: 		GetFileContents
		Purpose:		Opens the document specified by theFSSpec and puts
						the contents into theDocument.
	 -----------------------------------------------------------------------**/

#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

pascal OSErr GetFileContents(FSSpec spec, DPtr theDocument)
{
	long			theSize;
	short			oldRes;
	short			resFile;
	short			refNum;
	OSErr			err;
	HHandle		aHandle;
	Handle		gHandle;

	oldRes	=	CurResFile();
	resFile 	=	HOpenResFile(spec.vRefNum, spec.parID, spec.name, fsRdPerm);
	
	theDocument->u.reg.everLoaded = true;
	theDocument->u.reg.origFSSpec = spec;
	
	if (theDocument->inDataFork) {
#ifndef RUNTIME
		if (err = HOpenDF(spec.vRefNum, spec.parID, spec.name, fsRdPerm, &refNum)) {
			ShowError((StringPtr) "\pread file - HOpenDF", err);
#else
		if (err = HOpen(spec.vRefNum, spec.parID, spec.name, fsRdPerm, &refNum)) {
			ShowError((StringPtr) "\pread file - HOpen", err);
#endif
			refNum = 0;
			
			goto giveUp;
		}
	
		if (err = GetEOF(refNum, &theSize))
			goto giveUp;

  		gHandle = NewHandle(theSize);

		HLock(gHandle);
		if (err = FSRead(refNum, &theSize, *gHandle))
			return err;
		HUnlock(gHandle);
		FSClose(refNum);
	} else {
		gHandle	=	Get1NamedResource('TEXT', (StringPtr) "\p!");
		
		if (gHandle) 
			DetachResource(gHandle);
		else {
			err = ResError();
			
			goto giveUp;
		}
	}

	theDocument->u.reg.numSections = 0;

	if (resFile != -1) {
		aHandle = nil;

		if (Count1Resources('TFSS'))
			aHandle = (HHandle)Get1Resource('TFSS', 255);

		if (aHandle) {
			theDocument->u.reg.numSections = (*aHandle)->numSections;
			
			HLock((Handle) aHandle);
			ApplySettings(theDocument, *aHandle);
		}

		/*
			If there is a print record saved, ditch the old one
			created by new document and fill this one in
		*/
		if (Count1Resources('TFSP')) {
			if (theDocument->thePrintSetup)
				DisposHandle((Handle)theDocument->thePrintSetup);

			theDocument->thePrintSetup = (THPrint)Get1Resource('TFSP', 255);
		  	HandToHand((Handle *)&theDocument->thePrintSetup);

			PrValidate(theDocument->thePrintSetup);
		}

#ifndef RUNTIME
		if (theDocument->u.reg.numSections) {
			ReadSectionRecords(theDocument);
			ReadAllSectionResources(theDocument);
		}
#else
		theDocument->u.reg.numSections	=	0;
#endif

		CloseResFile(resFile);

		if (err = ResError()) {
			ShowError((StringPtr) "\pread file- CloseResFile", err);
			return err;
		}
	}

	HLock(gHandle);
	if (GetHandleSize(gHandle) > 32000) {
		PtrToXHand(
			*gHandle, 
			(*theDocument->theText)->hText, 
			GetHandleSize(gHandle));
		
		err = elvisErr;
	} else
		TESetText(*gHandle, GetHandleSize(gHandle), theDocument->theText);
	DisposHandle(gHandle);

	if (err == fnfErr)
		return noErr;
	else
  		return err;

giveUp:	
	if (refNum)
		FSClose(refNum);
		
	if (resFile > -1)
		CloseResFile(resFile);
	
	UseResFile(oldRes);
	
	return err;
} /* GetFileContents */


#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

pascal OSErr SaveAskingName(DPtr aDocument, Boolean canInteract)
{
	OSErr	myErr;

	if (aDocument->kind != kDocumentWindow || !aDocument->u.reg.everSaved) {

		if (canInteract) {
			if (myErr = GetFileNameToSaveAs(aDocument))
				return myErr;

			if (myErr = SaveWithoutTemp(aDocument, aDocument->theFSSpec))
				return myErr;			
#ifndef RUNTIME
			AssocAllSections(aDocument);
#endif

			return noErr;
		} else
			return errAENoUserInteraction;

	} else
		return SaveUsingTemp(aDocument);
}

#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

pascal OSErr SaveUsingTemp(DPtr theDocument)
{
	Str255	tempName;
	OSErr		err;
	FSSpec	tempFSSpec;

	/*save the file to disk using a temporary file*/
	/*this is the recommended way of doing things*/
	/*first write out the file to disk using a temporary filename*/
	/*if it is sucessfully written, exchange the temporary file with the last one saved*/
	/*then delete the temporary file- so if anything goes wrong, the original version is still there*/
	/*first generate the temporary filename*/

	GetTempFSSpec(theDocument, &tempFSSpec);

	if (err = DoCreate(tempFSSpec))
		return err;

	if (err = DoSave(theDocument, tempFSSpec))
		return err;
	
#ifndef RUNTIME
	if (err = FSpExchangeFiles(&tempFSSpec, &theDocument->theFSSpec))
		return err;
	
	/*we've exchanged the files, now delete the temporary one*/

	FSpDelete(&tempFSSpec);
#else
	err = FSpSmartMove(&tempFSSpec, &theDocument->theFSSpec);
#endif
	
	if (!err && theDocument->kind == kDocumentWindow) {
		theDocument->dirty 				= false;
		theDocument->u.reg.everSaved 	= true;
		theDocument->u.reg.origFSSpec = theDocument->theFSSpec;
	}

	return err;
}

pascal OSErr SaveWithoutTemp(DPtr theDocument, FSSpec spec)
{
	OSErr		err;
	
	if (err = DoSave(theDocument, spec))
		return err;

	if (theDocument->kind == kDocumentWindow) {
		theDocument->dirty 				= false;
		theDocument->u.reg.everSaved 	= true;
		theDocument->u.reg.origFSSpec = theDocument->theFSSpec = spec;
				
		SetWTitle(theDocument->theWindow, theDocument->theFSSpec.name);
	}
	
	return noErr;
}

#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

pascal short SaveScriptHook(short item, DialogPtr dlg, void * params)
{
	short				kind;
	ControlHandle	type;
	MenuHandle		typeMenu;
	Rect				r;
	DPtr				doc = (DPtr) params;
	
	if (GetWRefCon(dlg) != 'stdf')
		return item;
		
	switch (item) {
	case sfHookFirstCall:
		GetDItem(dlg, ssd_Type, &kind, (Handle *) &type, &r);
		typeMenu = (*(PopupPrivateDataHandle)(*type)->contrlData)->mHandle;
		
		AddExtensionsToMenu(typeMenu);
		SetControlMaximum(type, CountMItems(typeMenu));
		SetCtlValue(type, Type2Menu(doc->type));
		
		return sfHookFirstCall;
	case ssd_Type:
		GetDItem(dlg, item, &kind, (Handle *) &type, &r);
		
		doc->type = (DocType) Menu2Type(GetCtlValue(type));
		
		return sfHookNullEvent;
	default:
		return item;
	}
}

#if USESROUTINEDESCRIPTORS
RoutineDescriptor	uSaveScriptHook = 
		BUILD_ROUTINE_DESCRIPTOR(uppDlgHookYDProcInfo, SaveScriptHook);
#else
#define uSaveScriptHook *(DlgHookYDUPP)&SaveScriptHook
#endif

/*
	Fills in the document record with the user chosen destination
*/

pascal OSErr GetFileNameToSaveAs(DPtr theDocument)
{
#ifndef RUNTIME
	OSErr					err;
	StandardFileReply	reply;
	Str255            suggestName;
	Point					where;
	
	where.h = where.v = -1;

	GetWTitle(theDocument->theWindow, suggestName);

#ifndef RUNTIME
	HMSetDialogResID(SaveScriptDialog);
#endif
	CustomPutFile(
		(StringPtr) "\pSave Document As:", suggestName, &reply,
		SaveScriptDialog,
		where,
		&uSaveScriptHook,
		(ModalFilterYDUPP) nil,
		nil,
		(ActivateYDUPP) nil,
		theDocument);
#ifndef RUNTIME
	HMSetDialogResID(-1);
#endif

	if (reply.sfGood)
		switch (err = FSpDelete(&reply.sfFile)) {
		case noErr:
		case fnfErr:
			theDocument->theFSSpec = reply.sfFile;
			PLstrcpy(theDocument->theFileName, reply.sfFile.name);

			return noErr;
		default:
			return err;
		}
	else
		return userCanceledErr;
#else
	OSErr		err;
	SFReply	reply;
	Str255   suggestName;
	Point		where;

	where.h = where.v = 75;
	
	GetWTitle(theDocument->theWindow, suggestName);

	SFPutFile(
		where, (StringPtr) "\pSave Document As:", 
		suggestName, (DlgHookProcPtr) nil,
		&reply);
	
	if (reply.good) {
		FSSpec	spec;
		
		if (err = WD2FSSpec(reply.vRefNum, reply.fName, &spec))
			return err;
		switch (err = HDelete(spec.vRefNum, spec.parID, spec.name)) {
		case noErr:
		case fnfErr:
			theDocument->theFSSpec = spec;
			PLstrcpy(theDocument->theFileName, spec.name);

			return noErr;
		default:
			return err;
		}
	} else
		return userCanceledErr;
#endif
} /* GetFileNameToSaveAs */

#if !defined(powerc) && !defined(__powerc)
#pragma segment File
#endif

pascal DPtr MakeOldDoc(FSSpec aFSSpec, DocType type)
{
	DPtr  		theDocument;
	
	theDocument = NewDocument(true, kDocumentWindow);

	SetWTitle(theDocument->theWindow, aFSSpec.name);

	SetPort(theDocument->theWindow);

	theDocument->theFSSpec   = aFSSpec;

	PLstrcpy(theDocument->theFileName, aFSSpec.name);

	theDocument->dirty       		= false;
	theDocument->inDataFork			= type == kPlainTextDoc || type == kOldRuntime6Doc;
	
#ifndef RUNTIME
	/* We *can* open documents created by unknown save extensions, but we can't
	   save them again 
	*/
	if (CanSaveAs(type)) {
		theDocument->type 				= type;
		theDocument->u.reg.everSaved	= true;
	} else {
		theDocument->type					= (type == kOldRuntime6Doc) ? kRuntime6Doc : kPlainTextDoc;
		theDocument->u.reg.everSaved	= false;
	}
#else
	theDocument->type					= kPlainTextDoc;
	theDocument->u.reg.everSaved	= (type == kPlainTextDoc);
#endif

	return theDocument;
}

pascal OSErr OpenOld(FSSpec aFSSpec, DocType type)
{
	DPtr  		theDocument;
	OSErr 		fileErr;
	WindowPtr	win;

	if (win = AlreadyOpen(&aFSSpec, nil)) {
		SelectWindow(win);
		
		return noErr;
	}
	
	theDocument	=	MakeOldDoc(aFSSpec, type);
	
	fileErr = GetFileContents(aFSSpec, theDocument);

	if (!fileErr) {
		ResizeWindow(theDocument);
		DoShowWindow(theDocument->theWindow);
	} else {
		if (fileErr == elvisErr) {
			theDocument->u.reg.everSaved	= false;
#ifdef RUNTIME
			if (AEInteractWithUser(kAEDefaultTimeout, nil, nil))
#endif
				if (Alert(ElvisAlert, nil) == 1)
					SaveAskingName(theDocument, true);
		} else
			FileError((StringPtr) "\pError Opening ", aFSSpec.name);
		
		CloseMyWindow(theDocument->theWindow);
	}

	return fileErr;
} /* OpenOld */

pascal OSErr File2File(FSSpec aFSSpec, DocType type, FSSpec toFSSpec, DocType newtype)
{
	DPtr  		theDocument;
	OSErr 		fileErr;

	theDocument	=	MakeOldDoc(aFSSpec, type);
	
	switch (fileErr = GetFileContents(aFSSpec, theDocument)) {
	case elvisErr:
		fileErr = noErr;
		/* Fall through */
	case noErr:
		theDocument->type = newtype;
		theDocument->u.reg.everSaved	= false;
		if (SameFSSpec(&aFSSpec, &toFSSpec))
			fileErr = SaveUsingTemp(theDocument);
		else
			fileErr = SaveWithoutTemp(theDocument, toFSSpec);
		
		/* Fall through */
	default:
		CloseMyWindow(theDocument->theWindow);
	}

	return fileErr;
} /* File2File */

pascal OSErr Handle2File(Handle text, FSSpec toFSSpec, DocType newtype)
{
	DPtr  		theDocument;
	OSErr 		fileErr;

	theDocument = NewDocument(true, kDocumentWindow);
	
	HLock(text);
	PtrToXHand(*text, (*theDocument->theText)->hText, GetHandleSize(text));
	HUnlock(text);
	
	theDocument->u.reg.everSaved	= false;
	theDocument->type 				= newtype;

	fileErr = SaveWithoutTemp(theDocument, toFSSpec);
	
	CloseMyWindow(theDocument->theWindow);

	return fileErr;
} /* Handle2File */

pascal DocType GetDocTypeFromFile(short vRefNum, long dirID, StringPtr name)
{
	short			resFile;
	short			nuFile;
	OSType	**	rtType;
	DocType		type	=	kUnknownDoc;
	
	resFile	= CurResFile();
	nuFile	= HOpenResFile(vRefNum, dirID, name, fsRdPerm);
	
	if (nuFile != -1) {
		if (rtType = (OSType **) Get1Resource('MrPL', 128))
			type = **(DocType **) rtType;
			
		CloseResFile(nuFile);
	}
	UseResFile(resFile);
	
	return type;
}

pascal DocType GetDocTypeFromInfo(CInfoPBPtr info)
{
	DocType		type	=	kUnknownDoc;
	
	switch (info->hFileInfo.ioFlFndrInfo.fdType) {
	case 'APPL':
		if (info->hFileInfo.ioFlFndrInfo.fdCreator == MPAppSig) 
			/* A heuristic to separate old runtimes from PowerPC executables */
			if (info->hFileInfo.ioFlLgLen && info->hFileInfo.ioFlLgLen < 100000)
				return kOldRuntime6Doc; 
		
		break;
	case 'TEXT':
		return kPlainTextDoc;
	case 'pref':
		switch (info->hFileInfo.ioFlFndrInfo.fdCreator) {
		case MPAppSig:
		case MPRtSig:
			return kPreferenceDoc;
		}
		break;
	default:
		break;
	}
	
	/* Ultimately, with save extensions, every file could be ours */
	return GetDocTypeFromFile(
		info->hFileInfo.ioVRefNum, 
		info->hFileInfo.ioFlParID, 
		info->hFileInfo.ioNamePtr);;
}

pascal DocType GetDocType(FSSpec * spec)
{
	CInfoPBRec	info;

	if (FSpCatInfo(spec, &info))
		return kUnknownDoc;
	else
		return GetDocTypeFromInfo(&info);
}